print("import: 'sklearn'")
import sklearn

print("import: 'sklearn.cluster'")
import sklearn.cluster

print("import: 'sklearn.datasets'")
import sklearn.datasets

print("import: 'sklearn.decomposition'")
import sklearn.decomposition

print("import: 'sklearn.ensemble'")
import sklearn.ensemble

print("import: 'sklearn.feature_extraction'")
import sklearn.feature_extraction

print("import: 'sklearn.linear_model'")
import sklearn.linear_model

print("import: 'sklearn.manifold'")
import sklearn.manifold

print("import: 'sklearn.metrics'")
import sklearn.metrics

print("import: 'sklearn.neighbors'")
import sklearn.neighbors

print("import: 'sklearn.tree'")
import sklearn.tree

print("import: 'sklearn.utils'")
import sklearn.utils

print("import: 'sklearn.svm'")
import sklearn.svm

